/*
 * MiddlewareTest.cpp - CPPUnit test for libabiss middleware library
 *
 * Written 2004 by Erik Bos, Jennifer Blijlevens and Giel de Nijs
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <string>
#include <iostream>

#include "libabiss.h"

#include "MiddlewareTest.h"

using namespace std;

#define OPENRT_COUNT     4
#define WRITE_COUNT      500
#define PAGE_SIZE        4096
static const string abiss_fabiss_write_filename =
    string(getenv("ABISS_MOUNT_POINT")) + string("/fwr_test.bin");
static const string abiss_write_filename =
    string(getenv("ABISS_MOUNT_POINT")) + string("/wr_test.bin");

CPPUNIT_TEST_SUITE_REGISTRATION(MiddlewareTest);

string int2string(int value)
{
	char result[15];

	sprintf(result, "%d", value);
	return std::string(result);
}

MiddlewareTest::MiddlewareTest()
:  TestCase("MiddlewareTest")
{
	// fill buf with pattern '0x01 0x02 ...0xFF 0x01 ... 0xFF'
	for (int i = 0; i < DEFAULT_BUFSIZE; i++) {
		m_buf[i] = (i % 255) + 1;
	}
	m_buf[DEFAULT_BUFSIZE - 1] = '\0';
}

MiddlewareTest::~MiddlewareTest()
{
}

void MiddlewareTest::setUp()
{
	cerr << endl;
}

void MiddlewareTest::tearDown()
{
	cerr << endl;
}

void MiddlewareTest::test_abiss_failing_fopen()
{
	FILE *fp;

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/foo.bin");
	string mode = "r";
	int byterate = 1000000;
	int bufsize = DEFAULT_BUFSIZE;

	fp = abiss_fopen(path.c_str(), mode.c_str(), byterate, bufsize);
	CPPUNIT_ASSERT(NULL == fp);
}


void MiddlewareTest::test_abiss_fopen_fclose_byterate_OK()
{
	FILE *fp[OPENRT_COUNT];

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	string mode = "w+";
	int byterate = 1000000;
	int bufsize = DEFAULT_BUFSIZE;

	for (int i = 0; i < OPENRT_COUNT; i++) {
		fp[i] =
		    abiss_fopen(path.c_str(), mode.c_str(), byterate,
				bufsize);
		CPPUNIT_ASSERT(NULL != fp[i]);
	}

	for (int i = 0; i < OPENRT_COUNT; i++) {
		if (fp[i]) {
			CPPUNIT_ASSERT_EQUAL(0, abiss_fclose(fp[i]));
		}
	}
}


void MiddlewareTest::test_abiss_fopen_fclose_byterate_fail()
{
	FILE *fp;

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	string mode = "w+";
	int byterate = 100000000;
	int bufsize = DEFAULT_BUFSIZE;

	fp = abiss_fopen(path.c_str(), mode.c_str(), byterate, bufsize);
	CPPUNIT_ASSERT(NULL == fp);

	if (fp) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_fclose(fp));
	}
}


void MiddlewareTest::test_abiss_fopen_fclose_buffersize_OK()
{
	FILE *fp;

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	string mode = "w+";
	int byterate = 2000000;
	int bufsize = 10 * 256 * DEFAULT_BUFSIZE;

	fp = abiss_fopen(path.c_str(), mode.c_str(), byterate, bufsize);
	CPPUNIT_ASSERT(NULL != fp);

	if (fp) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_fclose(fp));
	}
}

void MiddlewareTest::test_abiss_fopen_fclose_buffersize_fail()
{
	FILE *fp;

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	string mode = "w+";
	int byterate = 2000000;
	int bufsize = 101 * 256 * DEFAULT_BUFSIZE;

	fp = abiss_fopen(path.c_str(), mode.c_str(), byterate, bufsize);
	CPPUNIT_ASSERT(NULL == fp);

	if (fp) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_fclose(fp));
	}
}

void MiddlewareTest::test_abiss_fread()
{
	FILE *fp;

	string mode = "r";
	int byterate = 1000000;
	char buf[DEFAULT_BUFSIZE];

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_fabiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}

	div_t div_result;

	if (filesize > -1) {
		div_result = div(filesize, DEFAULT_BUFSIZE);
	}

	int total_bytes_read = 0;
	int num_bytes_read;
	int i = 0;

	while (total_bytes_read != filesize) {
		num_bytes_read =
		    abiss_fread(buf, sizeof(char), DEFAULT_BUFSIZE, fp);
		total_bytes_read += num_bytes_read;

		CPPUNIT_ASSERT(num_bytes_read != -1);

		if (i == div_result.quot) {
			CPPUNIT_ASSERT_EQUAL(div_result.rem,
					     num_bytes_read);
		} else {
			CPPUNIT_ASSERT_EQUAL(DEFAULT_BUFSIZE,
					     num_bytes_read);

			int status = strncmp(m_buf, buf, DEFAULT_BUFSIZE);

			CPPUNIT_ASSERT_EQUAL(0, status);
		}
		i++;
	}

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}

void MiddlewareTest::test_abiss_fwrite()
{
	FILE *fp;

	string mode = "w";
	int byterate = 1000000;

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	int num_bytes_written = 0;

	for (int i = 0; i < WRITE_COUNT; i++) {
		num_bytes_written =
		    abiss_fwrite(m_buf, sizeof(char), DEFAULT_BUFSIZE, fp);
		CPPUNIT_ASSERT_EQUAL(DEFAULT_BUFSIZE, num_bytes_written);
	}

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_fabiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}
	CPPUNIT_ASSERT_EQUAL((WRITE_COUNT * DEFAULT_BUFSIZE), filesize);
}

void MiddlewareTest::test_abiss_fseek()
{
	FILE *fp;

	string mode = "r";
	int byterate = 1000000;
	char byte_under_seek;

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_fabiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}
	//////////////////////
	// SEEK_SET: offset is relative to the start of the file.
	//////////////////////
	long mbuf_offset = 7;
	long seek_offset = (5 * DEFAULT_BUFSIZE) + mbuf_offset;

	int error = abiss_fseek(fp, seek_offset, SEEK_SET);

	CPPUNIT_ASSERT_EQUAL(0, error);

	int num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);

	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_offset], byte_under_seek);

	//////////////////////
	// SEEK_CUR: offset is relative to current position indicator in the file.
	//////////////////////
	error = abiss_fseek(fp, seek_offset, SEEK_CUR);
	CPPUNIT_ASSERT_EQUAL(0, error);

	num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	ldiv_t div_result = ldiv((2 * seek_offset) + 1, DEFAULT_BUFSIZE);

	int mbuf_index = div_result.rem;

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	//////////////////////
	// SEEK_END: offset is relative to end-of-file.
	// NOTE: the offset is still added to the SEEK_END
	//////////////////////
	error = abiss_fseek(fp, -seek_offset, SEEK_END);
	CPPUNIT_ASSERT_EQUAL(0, error);

	num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	div_result = ldiv((filesize - seek_offset), DEFAULT_BUFSIZE);
	mbuf_index = div_result.rem;
//  mbuf_index = (filesize - seek_offset) % DEFAULT_BUFSIZE; DOES NOT WORK
	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}

void MiddlewareTest::test_abiss_fseeko()
{
	FILE *fp;

	string mode = "r";
	int byterate = 1000000;
	char byte_under_seek;

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_fabiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}
	//////////////////////
	// SEEK_SET: offset is relative to the start of the file.
	//////////////////////
	off_t mbuf_offset = 7;
	off_t seek_offset = (5 * DEFAULT_BUFSIZE) + mbuf_offset;

	int error = abiss_fseeko(fp, seek_offset, SEEK_SET);

	CPPUNIT_ASSERT_EQUAL(0, error);

	int num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);

	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_offset], byte_under_seek);

	//////////////////////
	// SEEK_CUR: offset is relative to current position indicator in the file.
	//////////////////////
	error = abiss_fseeko(fp, seek_offset, SEEK_CUR);
	CPPUNIT_ASSERT_EQUAL(0, error);

	num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	ldiv_t div_result = ldiv((2 * seek_offset) + 1, DEFAULT_BUFSIZE);

	int mbuf_index = div_result.rem;

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	//////////////////////
	// SEEK_END: offset is relative to end-of-file.
	// NOTE: the offset is still added to the SEEK_END
	//////////////////////
	error = abiss_fseeko(fp, -seek_offset, SEEK_END);
	CPPUNIT_ASSERT_EQUAL(0, error);

	num_bytes_read =
	    abiss_fread(&byte_under_seek, sizeof(char), 1, fp);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	div_result = ldiv((filesize - seek_offset), DEFAULT_BUFSIZE);
	mbuf_index = div_result.rem;
	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}

void MiddlewareTest::test_abiss_ftell()
{
	FILE *fp;

	string mode = "r";
	int byterate = 1000000;

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	long pos = ftell(fp);

	CPPUNIT_ASSERT(pos == 0);

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}

void MiddlewareTest::test_abiss_ftello()
{
	FILE *fp;

	string mode = "r";
	int byterate = 1000000;

	fp = abiss_fopen(abiss_fabiss_write_filename.c_str(), mode.c_str(),
			 byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(NULL != fp);

	off_t pos = ftell(fp);

	CPPUNIT_ASSERT(pos == 0);

	if (fp) {
		int status = abiss_fclose(fp);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}



void MiddlewareTest::test_abiss_failing_open()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;
	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/foo.bin");
	int byterate = 1000000;
	int bufsize = DEFAULT_BUFSIZE;

	fd = abiss_open(path.c_str(), flags, mode, byterate, bufsize);
	CPPUNIT_ASSERT(-1 == fd);
}


void MiddlewareTest::test_abiss_open_close_byterate_OK()
{
	int fd[OPENRT_COUNT];
	int flags = O_RDONLY;

	mode_t mode = 0;
	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	int byterate = 1000000;
	int bufsize = DEFAULT_BUFSIZE;

	for (int i = 0; i < OPENRT_COUNT; i++) {
		fd[i] =
		    abiss_open(path.c_str(), flags, mode, byterate,
			       bufsize);
		CPPUNIT_ASSERT(-1 != fd[i]);
	}

	for (int i = 0; i < OPENRT_COUNT; i++) {
		if (-1 != fd[i]) {
			CPPUNIT_ASSERT_EQUAL(0, abiss_close(fd[i]));
		}
	}
}


void MiddlewareTest::test_abiss_open_close_byterate_fail()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;
	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	int byterate = 100000000;
	int bufsize = DEFAULT_BUFSIZE;

	fd = abiss_open(path.c_str(), flags, mode, byterate, bufsize);
	CPPUNIT_ASSERT(-1 == fd);

	if (-1 != fd) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_close(fd));
	}
}


void MiddlewareTest::test_abiss_open_close_buffersize_OK()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;
	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	int byterate = 2000000;
	int bufsize = 256 * DEFAULT_BUFSIZE;

	fd = abiss_open(path.c_str(), flags, mode, byterate, bufsize);
	CPPUNIT_ASSERT(-1 != fd);

	if (-1 != fd) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_close(fd));
	}
}

void MiddlewareTest::test_abiss_open_close_buffersize_fail()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;

	string path =
	    string(getenv("ABISS_MOUNT_POINT")) + string("/test.bin");
	int byterate = 2000000;
	int bufsize = 101 * 256 * DEFAULT_BUFSIZE;

	fd = abiss_open(path.c_str(), flags, mode, byterate, bufsize);
	CPPUNIT_ASSERT(-1 == fd);

	if (-1 != fd) {
		CPPUNIT_ASSERT_EQUAL(0, abiss_close(fd));
	}
}

void MiddlewareTest::test_abiss_read()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;
	int byterate = 1000000;
	char buf[DEFAULT_BUFSIZE];

	fd = abiss_open(abiss_write_filename.c_str(), flags, mode,
			byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(-1 != fd);

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}

	int total_bytes_read = 0;
	int num_bytes_read;
	int i = 0;

	while (total_bytes_read != filesize) {
		num_bytes_read = abiss_read(fd, buf, DEFAULT_BUFSIZE);
		total_bytes_read += num_bytes_read;
		CPPUNIT_ASSERT(num_bytes_read != -1);
		CPPUNIT_ASSERT(num_bytes_read <= DEFAULT_BUFSIZE);
		i++;
	}

	if (fd != -1) {
		int status = abiss_close(fd);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}
}

void MiddlewareTest::test_abiss_write()
{
	int fd;
	int flags = O_RDWR | O_CREAT | O_TRUNC;

	mode_t mode = S_IRWXU;
	int byterate = 1000000;

	fd = abiss_open(abiss_write_filename.c_str(), flags, mode,
			byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(-1 != fd);

	int num_bytes_written = 0;

	for (int i = 0; i < WRITE_COUNT; i++) {
		num_bytes_written =
		    abiss_write(fd, m_buf, DEFAULT_BUFSIZE);
		CPPUNIT_ASSERT_EQUAL(DEFAULT_BUFSIZE, num_bytes_written);
	}

	if (fd != -1) {
		int status = abiss_close(fd);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}
	CPPUNIT_ASSERT_EQUAL((WRITE_COUNT * DEFAULT_BUFSIZE), filesize);
}

void MiddlewareTest::test_abiss_lseek()
{
	int fd;
	int flags = O_RDONLY;

	mode_t mode = 0;
	int byterate = 1000000;
	char byte_under_seek;

	fd = abiss_open(abiss_write_filename.c_str(), flags, mode,
			byterate, DEFAULT_BUFSIZE);
	CPPUNIT_ASSERT(-1 != fd);

	int filesize = -1;
	struct stat filestat;

	if (stat(abiss_write_filename.c_str(), &filestat) == 0) {
		filesize = (int) filestat.st_size;
	}
	//////////////////////
	// SEEK_SET: offset is relative to the start of the file.
	//////////////////////
	off_t mbuf_offset = 7;
	off_t seek_offset = (5 * DEFAULT_BUFSIZE) + mbuf_offset;

	off_t new_pos = abiss_lseek(fd, seek_offset, SEEK_SET);

	CPPUNIT_ASSERT_EQUAL(seek_offset, new_pos);

	int num_bytes_read = abiss_read(fd, &byte_under_seek, 1);

	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_offset], byte_under_seek);

	sleep(3);		// wait until the playout point has catched up
	off_t playout_point = abiss_get_playoutpoint_fd(fd);

	new_pos = new_pos & ~(PAGE_SIZE - 1);

	CPPUNIT_ASSERT(new_pos == playout_point);

	//////////////////////
	// SEEK_CUR: offset is relative to current position indicator in the file.
	//////////////////////
	new_pos = abiss_lseek(fd, seek_offset, SEEK_CUR);
	CPPUNIT_ASSERT_EQUAL(((2 * seek_offset) + 1), new_pos);

	num_bytes_read = abiss_read(fd, &byte_under_seek, 1);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	ldiv_t div_result = ldiv((2 * seek_offset) + 1, DEFAULT_BUFSIZE);

	int mbuf_index = div_result.rem;

	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	//////////////////////
	// SEEK_END: offset is relative to end-of-file.
	// NOTE: the offset is still added to the SEEK_END
	//////////////////////
	new_pos = abiss_lseek(fd, -seek_offset, SEEK_END);
	CPPUNIT_ASSERT_EQUAL((filesize - seek_offset), new_pos);

	num_bytes_read = abiss_read(fd, &byte_under_seek, 1);
	CPPUNIT_ASSERT_EQUAL(1, num_bytes_read);

	div_result = ldiv((filesize - seek_offset), DEFAULT_BUFSIZE);
	mbuf_index = div_result.rem;
	CPPUNIT_ASSERT_EQUAL(m_buf[mbuf_index], byte_under_seek);

	if (fd != -1) {
		int status = abiss_close(fd);

		CPPUNIT_ASSERT_EQUAL(0, status);
	}

}
