/*
 * racc.c - Random reader/writer
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */


#define _XOPEN_SOURCE 500

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mount.h>

#include "racc.h"


#define DEFAULT_STEP_SIZE 512


int verbose = 0;
int synchronize = 0;


void show_time(const char *fmt,...)
{
    static struct timeval then;
    struct timeval now;

    if (gettimeofday(&now,NULL) < 0) {
	perror("gettimeofday");
	exit(1);
    }
    if (fmt) {
	va_list ap;
	long long delta;

	delta = (now.tv_sec-then.tv_sec)*1000000LL;
	delta += now.tv_usec-then.tv_usec;
	va_start(ap,fmt);
	vfprintf(stderr,fmt,ap);
	fprintf(stderr,": %lld.%06lld\n",delta/1000000,delta % 1000000);
	va_end(ap);
    }
    then = now;
}


static void usage(const char *name)
{
    fprintf(stderr,
"usage: %s [-d] "
#ifdef HAVE_LIBAIO
  "[-l] "
#endif
  "[-n access_size] [-s] [-v] [-w|-W]\n"
"           file_or_device accesses [step_size]\n\n"
"  -d              use direct IO (O_DIRECT)\n"
#ifdef HAVE_LIBAIO
"  -l              use Linux-specific libaio instead of POSIX AIO\n"
#endif
"  -n access_size  transfer that many bytes in each read or write access\n"
"                  (default: 1 byte)\n"
"  -s              synchronize file data to disk before terminating\n"
"  -v              verbose: report what we're doing (to standard error)\n"
"  -w              write non-destructively (read, then write)\n"
"  -W              write destructively (write zeroes)\n\n"
"  file_or_device  the file or the block device to access\n"
"  accesses        the number of (read-only, read-write, or write-only)\n"
"                  accesses\n"
"  step_size       the granularity of locations accessed, in bytes\n"
"                  (default: %d)\n"
      ,name,DEFAULT_STEP_SIZE);
    exit(1);
}


int main(int argc,char **argv)
{
    char *end;
    int flags = O_RDONLY;
#ifdef HAVE_LIBAIO
    int use_libaio = 0;
#endif
    int o_direct = 0;
    int access_size = 1,step_size = DEFAULT_STEP_SIZE,accesses,locations;
    off_t obj_size;
    int c,fd;
    struct stat st;

    show_time(NULL);
    srandom(time(NULL));
    while ((c = getopt(argc,argv,"dln:svwW")) != EOF)
	switch (c) {
	    case 'd':
		o_direct = O_DIRECT;
		break;
	    case 'l':
#ifdef HAVE_LIBAIO
		use_libaio = 1;
		break;
#else /* HAVE_LIBAIO */
		fprintf(stderr,"%s was compiled without libaio support\n",
		  *argv);
		return 1;
#endif /* !HAVE_LIBAIO */
	    case 'n':
		access_size = strtoul(optarg,&end,0);
		if (*end)
		    usage(*argv);
		break;
	    case 's':
		synchronize = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'w':
		flags = O_RDWR;
		break;
	    case 'W':
		flags = O_WRONLY;
		break;
	    default:
		usage(*argv);
	}
    switch (argc-optind) {
	case 3:
	    step_size = strtoul(argv[optind+2],&end,0);
	    if (*end)
		usage(*argv);
	    /* fall through */
	case 2:
	    fd = open(argv[optind],flags | o_direct);
	    if (fd < 0) {
		perror(argv[optind]);
		return 1;
	    }
	    accesses = strtoul(argv[optind+1],&end,0);
	    if (*end)
		usage(*argv);
	    break;
	default:
	    usage(*argv);
    }
    if (fstat(fd,&st) < 0) {
	perror("fstat");
	return 1;
    }
    if (S_ISREG(st.st_mode))
	obj_size = st.st_size;
    else if (S_ISBLK(st.st_mode)) {
	unsigned long sectors;

	if (ioctl(fd,BLKGETSIZE,&sectors) < 0) {
	    perror("ioctl BLKGETSIZE");
	    return 1;
	}
	obj_size = (off_t) sectors << 9;
    }
    else {
	fprintf(stderr,"%s is neither a regular file nor a block device\n",
	  argv[optind]);
	return 1;
    }
    locations =
       obj_size/step_size > INT_MAX ? INT_MAX : obj_size/step_size;
    if (verbose) {
	const char *access_type[] = {
	    [O_RDONLY] = "reads from",
	    [O_RDWR] = "rewrites of",
	    [O_WRONLY] = "writes to",
	};

	fprintf(stderr,
	  "%s: %d %s %d %d-byte locations in %llu bytes\n",
	  argv[optind],accesses,access_type[flags],
	  locations,step_size,(unsigned long long) obj_size);
    }
#ifdef HAVE_LIBAIO
    if (use_libaio)
	racc_linux(fd,flags,obj_size,access_size,step_size,accesses,locations);
    else
#endif /* HAVE_LIBAIO */
	racc_posix(fd,flags,obj_size,access_size,step_size,accesses,locations);

    if (synchronize) {
	if (fdatasync(fd) < 0) {
	    perror("fdatasync");
	    exit(1);
	}
	if (verbose)
	    show_time("sync");
    }
    return 0;
}
