/*
 * rdrt.c - Real-time reader/writer
 *
 * Written 2004,2005 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004,2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.GPLv2
 */


#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <abiss_msg.h>
#include <abiss_msg_basic.h>

#include "seq.h"


#define TIMESTAMPS_MS		8
#define TIMESTAMPS_US		4
#define BARS			78
#define MAX_MLOCKED_READS	100000 /* don't try to mlock lots of memory */

struct stats {
    int dev_us;			/* deviation (in us) */
    int lat_us;			/* read/write latency (us) */
    int pos_us;			/* positioning latency (us) */
    int mm_us;			/* memory reclaim latency (us) */
    long long int run_us;	/* run time (us) */
};

enum ps_mode {
    ps_jitter,			/* read/write jitter */
    ps_lat,			/* read/write latency */
    ps_pos,			/* positioning latency */
    ps_mm,			/* memory reclaim latency */
};


static int unit_us = 1000; /* output unit; 1000 us */
static int timestamps = TIMESTAMPS_MS;


static void usage(const char *name)
{
    fprintf(stderr,
      "usage: %s [-v] [-s] [-o] [-x] [-M] [-m|-u]\n"
      "            [-t|-h|-c|-r|-l|-p|-a|-d|-q]\n"
      "            [-B bin_width] [-i increment [-f first_offset]\n"
      "            [-w word_size] [-b start_value] [-e end_value]] file_name\n"
      "            [rate] read/write_size [read_ahead_bytes"
      "]\n\n"
      "  rate              read/write and fill rate (bytes per second)\n"
      "  read/write_size   read/write size (bytes)\n"
      "  read_ahead_bytes  read-ahead buffer (bytes)\n"
      "\n"
      "Sizes and rates can be followed by one of the SI prefixes \"k\" (1000)\n"
      "\"M\" (1000000), or \"G\" (1000000000). Also the respective unit\n"
      "(\"B\" or \"Bps\") can be specified, but is optional.\n"
      "\n"
      "General options:\n"
      "  -o   use SCHED_OTHER, i.e. don't raise CPU priority (and don't "
      "mlockall)\n"
      "  -x   allow excessive resource use (which may kill the system)\n"
      "  -v   verbose: print what will be, and what has been done\n"
      "  -s   silent: don't print progress indications\n"
      "  -M   enable direct reclaiming of free memory\n"
      "Statistics options:\n"
      "  -t   print jitter over time\n"
      "  -h   print histogram for jitter\n"
      "  -c   like -h, but cumulative\n"
      "  -r   print read/write latency over time\n"
      "  -l   print histogram for read/write latency\n"
      "  -p   print playout point advancing latency over time\n"
      "  -a   print histogram for playout point advancing latency\n"
      "  -d   print direct memory reclaim duration over time\n"
      "  -q   print histogram for memory reclaim duration\n"
      "  -m   use milliseconds (default)\n"
      "  -u   use microseconds instead of milliseconds\n"
      "  -w bin_width\n"
      "       set the histogram bin width (in microseconds)\n"
      "Verification options:\n"
      "  -i increment\n"
      "       verify every \"increment\"-th byte\n"
      "  -f first_offset\n"
      "       begin verification at \"start_offset\" bytes. Default is 0.\n"
      "  -b start_value\n"
      "       begin counting at \"start_value\". Default is 0.\n"
      "  -e end_value\n"
      "       wrap back to \"start_value\" after reaching \"end_value\".\n"
      "       Must be equal or greater than \"start_value\" and must fit in\n"
      "       the word size. Default is 255.\n"
      ,name);
    exit(1);
}


static int si(const char *name,const char *end,const char *unit)
{
    if (!*end)
	return 1;
    if (!strcmp(end,unit))
	return 1;
    if (end[1] && strcmp(end+1,unit))
	usage(name);
    switch (*end) {
	case 'k':
	    return 1000;
	case 'M':
	    return 1000*1000;
	case 'G':
	    return 1000*1000*1000;
	default:
	    usage(name);
	    return 0; /* stupid gcc */
    }
}


static void do_gettimeofday(struct timeval *tv)
{
    if (gettimeofday(tv,NULL) < 0) {
	perror("gettimeofday");
	exit(1);
    }
}


static void tv_add_us(struct timeval *a,int b)
{
    a->tv_usec += b;
    while (a->tv_usec >= 1000000) {
	a->tv_usec -= 1000000;
	a->tv_sec++;
    }
}


static void tv_sub(struct timeval *a,const struct timeval *b)
{
    a->tv_sec -= b->tv_sec;
    a->tv_usec -= b->tv_usec;
    if (a->tv_usec >= 0)
	return;
    a->tv_usec += 1000000;
    a->tv_sec--;
}


static void tv_sub_neg(struct timeval *a,const struct timeval *b)
{
    a->tv_sec = b->tv_sec-a->tv_sec;
    a->tv_usec = b->tv_usec-a->tv_usec;
    if (a->tv_usec >= 0)
	return;
    a->tv_usec += 1000000;
    a->tv_sec--;
}


static int cmp_dev(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->dev_us < b->dev_us ? -1 :
      a->dev_us > b->dev_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_lat(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->lat_us < b->lat_us ? -1 :
      a->lat_us > b->lat_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_pos(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->pos_us < b->pos_us ? -1 :
      a->pos_us > b->pos_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_mm(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->mm_us < b->mm_us ? -1 :
      a->mm_us > b->mm_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static void print_stat(const struct stats *s,int n,int cumul,int sum,
  enum ps_mode mode)
{
    int i,x;

    switch (mode) {
	case ps_jitter:
	    x = s->dev_us;
	    break;
	case ps_lat:
	    x = s->lat_us;
	    break;
	case ps_pos:
	    x = s->pos_us;
	    break;
	case ps_mm:
	    x = s->mm_us;
	    break;
	default:
	    abort();
    }
    printf("%5d %6d  # ",x/unit_us,cumul ? sum : n);
    for (i = 0; i != timestamps && i != n; i++)
	printf("%lld%s",s[i].run_us/unit_us,i != n-1 ? "," : "");
    if (i < n-1)
	printf("...,");
    if (i != n)
	printf("%lld",s[n-1].run_us/unit_us);
    printf("\n");
}


static void realtimize(void)
{
    struct sched_param prm;

    if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
	perror("mlockall");
	exit(1);
    }
    prm.sched_priority = sched_get_priority_max(SCHED_FIFO);
    if (prm.sched_priority < 0) {
	perror("sched_get_priority_max SCHED_FIFO");
	exit(1);
    }
    if (sched_setscheduler(0,SCHED_FIFO,&prm) < 0) {
	perror("sched_setscheduler SCHED_FIFO");
	exit(1);
    }
}


static void unrealtime(void)
{
    struct sched_param prm = { .sched_priority = 0 };

    if (sched_setscheduler(0,SCHED_OTHER,&prm) < 0) {
	perror("sched_setscheduler SCHED_OTHER");
	exit(1);
    }
    if (munlockall() < 0) {
	perror("munlockall");
	exit(1);
    }
}


int main(int argc,char **argv)
{
    static struct abiss_sched_basic_prm basic_prm;
    static struct abiss_attach_msg attach_msg = {
	.header = {
	    .type = abiss_attach,
	},
    };
    static struct abiss_position_msg pos_msg = {
	.header = {
	    .type = abiss_position,
	},
	.whence = SEEK_CUR,
	.pos = 0,
    };
    static struct timeval now;

    const char *name = *argv;
    struct stat st;
    struct timeval start,goal;
    int stat_time = 0,stat_hist = 0,stat_cumul = 0,stat_lat_time = 0,
      stat_lat_hist = 0,stat_pos_time = 0,stat_pos_hist = 0;
    int stat_mm_time = 0,stat_mm_hist = 0;
    int unit_prefix = 'm';
    int bin_us = 1000; /* 1 ms histogram bin width */
    int verbose = 0,silent = 0,rt,raise_prio = 1,excessive = 0;
    int increment = 0,offset = 0,start_value = 0,end_value = 255,word_size = 1;
    int do_write = 0,sequence = 0,num_writes = -1;
    int read_size;
    int *fill_Bps;
    char *end;
    void *buffer;
    int fd,interval_us,n = 0,num_reads,bars = 0,value;
    int seq_errors = 0;
    int direct_mm = 0;
    struct stats *stat;
    int c;

    /* lower-case options still available: gjkyz */
    while ((c = getopt(argc,argv,"ab:cde:f:hi:lmn:opqrstuvw:xMSW")) != EOF)
	switch (c) {
	    case 't':
		stat_time = 1;
		break;
	    case 'h':
		stat_hist = 1;
		break;
	    case 'c':
		stat_cumul = 1;
		break;
	    case 'r':
		stat_lat_time = 1;
		break;
	    case 'l':
		stat_lat_hist = 1;
		break;
	    case 'o':
		raise_prio = 0;
		break;
	    case 'x':
		excessive = 1;
		break;
	    case 'p':
		stat_pos_time = 1;
		break;
	    case 'a':
		stat_pos_hist = 1;
		break;
	    case 's':
		silent = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'i':
		increment = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'f':
		offset = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'b':
		start_value = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'e':
		end_value = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'm':
		unit_us = 1000;
		unit_prefix = 'm';
		timestamps = TIMESTAMPS_MS;
		break;
	    case 'u':
		unit_us = 1;
		unit_prefix = 'u';
		timestamps = TIMESTAMPS_US;
		break;
	    case 'w':
		bin_us = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'n':
		num_writes = strtoul(optarg,&end,10);
		break;
	    case 'M':
		direct_mm = 1;
		break;
	    case 'd':
		stat_mm_time = 1;
		break;
	    case 'q':
		stat_mm_hist = 1;
		break;
	    case 'W':
		do_write = 1;
		break;
	    case 'S':
		sequence = 1;
		break;
	    default:
		usage(name);
	}

    if (increment && (increment < word_size || start_value > end_value ||
      end_value >> word_size*8))
	usage(name);
    value = start_value;

    switch (argc-optind) {
	case 4:
	    rt = 1;
	    strcpy(attach_msg.sched_name,"basic");
	    attach_msg.sched_prm = &basic_prm;
	    basic_prm.ra_bytes = strtoul(argv[optind+3],&end,0);
	    basic_prm.ra_bytes *= si(name,end,"B");
	    fill_Bps = &basic_prm.fill_Bps;
	    break;
	case 2:
	    rt = 0;
	    fill_Bps = NULL;
	    break;
	case 3:
	    rt = 0;
	    fill_Bps = &basic_prm.fill_Bps; /* pick any */
	    break;
	default:
	    usage(name);
    }

    if (bin_us < unit_us)
	fprintf(stderr,"warning: bin width (%d us) < output unit (%d us)\n",
	  bin_us,unit_us);

    fd = open(argv[optind],
      do_write ? O_WRONLY | (num_writes == -1 ? 0 : O_CREAT): O_RDONLY,0666);
    if (fd < 0) {
	perror(argv[optind]);
	return 1;
    }
    if (fstat(fd,&st) < 0) {
	perror("fstat");
	return 1;
    }
    seq = st.st_ino;

    read_size = strtoul(argv[optind+1+!!fill_Bps],&end,0);
    read_size *= si(name,end,"B");
    if (read_size <= 0) {
	fprintf(stderr,"\"read/write_size\" must be greater than zero\n");
	return 1;
    }

    buffer = malloc(read_size);
    if (!buffer) {
	perror("malloc");
	return 1;
    }
    if (do_write)
	memset(buffer,'U',read_size);

    num_reads = (st.st_size+read_size-1)/read_size;
    if (num_writes != -1) {
	num_reads = num_writes;
	st.st_size = num_writes*read_size;
    }

    if (raise_prio && num_reads > MAX_MLOCKED_READS && !excessive) {
	fprintf(stderr,"won't mlockall with %d > %d reads/writes "
          "(please use -o or -x)\n",num_reads,MAX_MLOCKED_READS);
	return 0;
    }

    stat = malloc(num_reads*sizeof(struct stats));
    if (!stat) {
	perror("malloc");
	return 1;
    }

    if (fill_Bps) {
	*fill_Bps = strtoul(argv[optind+1],&end,0);
	*fill_Bps *= si(name,end,"Bps");
	if (*fill_Bps <= 0) {
	    fprintf(stderr,"\"rate\" must be greater than zero\n");
	    return 1;
	}
	interval_us = read_size*1000000.0/ *fill_Bps;
                                       /* ^___________ mind the gap :-) */
	if (verbose) {
	    double t = num_reads ? (num_reads-1.0)*interval_us/1000000.0 : 0;

	    printf("# %s %lu bytes (%d %s operation%s) in %.6f s\n",
	      do_write ? "writing" : "reading",(unsigned long) st.st_size,
	      num_reads,do_write ? "write" : "read",num_reads == 1 ? "" : "s",
	      t);
	    if (t)
		printf("# (%f bytes/sec, interval %d us)\n",
		  st.st_size/t,interval_us);
	}
    }
    else {
	if (verbose)
	    printf("# greedily %s %lu bytes (%d %s operation%s)\n",
	      do_write ? "writing" : "reading",(unsigned long) st.st_size,
	      num_reads,do_write ? "write" : "read",num_reads == 1 ? "" : "s");
    }

    if (rt && ioctl(fd,ABISS_IOCTL,&attach_msg) < 0) {
	perror("ioctl ABISS_IOCTL (abiss_attach)");
	return 1;
    }
    if (raise_prio)
	realtimize();
    do_gettimeofday(&start);
    goal = start;

    while (1) {
	struct timeval before_read,after_read,after_pos;
	struct timespec ts;
	int got,dt,sec;

	do_gettimeofday(&before_read);
	if (do_write) {
	    size_t size = st.st_size > read_size ? read_size : st.st_size;

	    if (sequence)
		put_sequence(buffer,size);
	    got = write(fd,buffer,size);
	    if (got < 0) {
		perror("write");
		return 1;
	    }
	}
	else {
	    got = read(fd,buffer,read_size);
	    if (got < 0) {
		perror("read");
		return 1;
	    }
	    if (sequence) {
		if (check_sequence(buffer,got)) {
		    fprintf(stderr,"sequence error in read #%d\n",n);
		    seq_errors++;
		}
	    }
	}
	do_gettimeofday(&after_read);

	if (rt && ioctl(fd,ABISS_IOCTL,&pos_msg) < 0) {
	    perror("ioctl ABISS_IOCTL (abiss_position)");
	    return 1;
	}
	do_gettimeofday(&after_pos);

	dt = (after_read.tv_sec-goal.tv_sec)*1000000+
	  after_read.tv_usec-goal.tv_usec;
	stat[n].dev_us = dt;

	if (!fill_Bps)
	    goal = after_read;

	tv_sub(&after_pos,&after_read);
	stat[n].pos_us = after_pos.tv_sec*1000000+after_pos.tv_usec;

	tv_sub(&after_read,&before_read);
	stat[n].lat_us = after_read.tv_sec*1000000+after_read.tv_usec;

	tv_sub(&before_read,&start);
	stat[n].run_us = before_read.tv_sec*1000000LLU+before_read.tv_usec;

	n++;

	if (increment) {
	    while (offset < got) {
		if (((uint8_t *) buffer)[offset] != value) {
		    fprintf(stderr,"invalid value at offset %ld: expected %u,"
		      " got %d\n",(n-1L)*read_size+offset,value,
		      ((uint8_t *) buffer)[offset]);
		    return 1;
		}
		value = value == end_value ? start_value : value+1;
		offset += increment;
	    }
	    offset -= got;
	}

	if (!silent)
	    while (bars != (num_reads ? n*BARS/num_reads : BARS)) {
		write(2,"#",1);
		bars++;
	    }

	do_gettimeofday(&now);

	st.st_size -= got;
	if (!st.st_size)
	    break;
	if (got < read_size) {
	    fprintf(stderr,"%s(%d) returned %d, with supposedly %lu left\n",
	      do_write ? "write" : "read",read_size,got,
	      (unsigned long) st.st_size+got);
	    return 1;
	}

	if (!fill_Bps)
	    continue;

	if (direct_mm) {
		static struct abiss_msg msg = {
			.type = abiss_direct_mm,
		};
		struct timeval then = now;

		if (ioctl(fd,ABISS_IOCTL,&msg) < 0) {
			perror("ioctl ABISS_IOCTL (abiss_direct_mm)");
			exit(1);
		}
		do_gettimeofday(&now);
		tv_sub_neg(&then,&now);
		stat[n].mm_us = then.tv_sec*1000000LLU+then.tv_usec;
	}

	tv_add_us(&goal,interval_us);
	sec = goal.tv_sec-now.tv_sec;
	ts.tv_nsec = (goal.tv_usec-now.tv_usec)*1000;
	if (ts.tv_nsec < 0) {
		ts.tv_nsec += 1000*1000*1000;
		sec--;
	}
	if (sec < 0)
	    goal = now;
	else {
	    ts.tv_sec = sec;
	    if (nanosleep(&ts,NULL) < 0) {
		perror("nanosleep");
		return 1;
	    }
	}
    }
    if (raise_prio)
	unrealtime();
    if (!silent)
	write(2,"\n",1);

    if (verbose) {
	tv_sub(&now,&start);
	printf("# completed after %d.%06d seconds\n",
	  (int) now.tv_sec,(int) now.tv_usec);
    }
    if (stat_time) {
	int i;

	printf("# Time of sample (%cs), jitter (%cs)\n",
	  unit_prefix,unit_prefix);
	for (i = 0; i != n; i++)
	    printf("%lld %d\n",stat[i].run_us/unit_us,stat[i].dev_us/unit_us);
	printf("\n");
    }
    if (stat_lat_time) {
	int i;

	printf("# Read latency (%cs), over time (%cs)\n",
	  unit_prefix,unit_prefix);
	for (i = 0; i != n; i++)
	    printf("%lld %d\n",stat[i].run_us/unit_us,stat[i].lat_us/unit_us);
	printf("\n");
    }
    if (stat_pos_time) {
	int i;

	printf("# Playout point advancing latency (%cs), over time (%cs)\n",
	  unit_prefix,unit_prefix);
	for (i = 0; i != n; i++)
	    printf("%lld %d\n",stat[i].run_us/unit_us,stat[i].pos_us/unit_us);
	printf("\n");
    }
    if (stat_mm_time) {
	int i;

	printf("# Direct memory reclaim duration (%cs), over time (%cs)\n",
	  unit_prefix,unit_prefix);
	for (i = 0; i != n; i++)
	    printf("%lld %d\n",stat[i].run_us/unit_us,stat[i].mm_us/unit_us);
	printf("\n");
    }
    if (stat_hist || stat_cumul)
	qsort(stat,n,sizeof(struct stats),cmp_dev);
    if (stat_hist) {
	int cnt = 0,i;

	printf("# Jitter (%cs), number of occurrences (per %d us)\n",
	  unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].dev_us/bin_us != stat[i-1].dev_us/bin_us) {
		print_stat(stat+i-cnt,cnt,0,0,ps_jitter);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat(stat+i-cnt,cnt,0,0,ps_jitter);
	printf("\n");
    }
    if (stat_cumul) {
	int cnt = 0,i,sum = 0;

	printf(
	  "# Cumulative: jitter (%cs), number of occurrences (per %d us)\n",
	  unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].dev_us/bin_us != stat[i-1].dev_us/bin_us) {
		print_stat(stat+i-cnt,cnt,1,sum,ps_jitter);
		cnt = 0;
	    }
	    sum++;
	    cnt++;
	}
	print_stat(stat+i-cnt,cnt,1,sum,ps_jitter);
	printf("\n");
    }
    if (stat_lat_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_lat);
	printf("# Latency for %s (%cs), number of occurrences (per %d us)\n",
	  do_write ? "write" : "read",unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].lat_us/bin_us != stat[i-1].lat_us/bin_us) {
		print_stat(stat+i-cnt,cnt,0,0,ps_lat);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat(stat+i-cnt,cnt,0,0,ps_lat);
	printf("\n");
    }
    if (stat_pos_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_pos);
	printf("# Advancing (playout point) latency (%cs), "
	  "number of occurrences (per %d us)\n",unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].pos_us/bin_us != stat[i-1].pos_us/bin_us) {
		print_stat(stat+i-cnt,cnt,0,0,ps_pos);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat(stat+i-cnt,cnt,0,0,ps_pos);
	printf("\n");
    }
    if (stat_mm_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_mm);
	printf("# Reclaim (free memory) latency (%cs), "
	  "number of occurrences (per %d us)\n",unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].mm_us/bin_us != stat[i-1].mm_us/bin_us) {
		print_stat(stat+i-cnt,cnt,0,0,ps_mm);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat(stat+i-cnt,cnt,0,0,ps_mm);
	printf("\n");
    }
    if (seq_errors) {
	fprintf(stderr,"%d sequence error%s\n",seq_errors,
	  seq_errors == 1 ? "" : "s");
	return 1;
    }
    return 0;
}
